#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = (
    "config_def", "guerilla_workspace_def", "v2_config_def",
    "upgrade_config"
    )

from mozzarilla.defs import config_def, guerilla_workspace_def,\
     v2_config_def, upgrade_config
