#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# ##############
#   metadata   #
# ##############
__author__ = "Devin Bobadilla, Michelle van der Graaf"
#           YYYY.MM.DD
__date__ = "2020.02.29"
__version__ = (1, 9, 0)
__website__ = "https://github.com/Sigmmma/mozzarilla"
